# Generated by Django 3.2.19 on 2023-06-05 11:02

import django.db.models.deletion
from django.db import migrations, models

import power_query.json


class Migration(migrations.Migration):
    dependencies = [
        ("power_query", "0013_migration"),
    ]

    operations = [
        migrations.AddField(
            model_name="parametrizer",
            name="source",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="power_query.query",
            ),
        ),
        migrations.AddField(
            model_name="query",
            name="celery_task",
            field=models.CharField(blank=True, max_length=36, null=True),
        ),
        migrations.AddField(
            model_name="report",
            name="celery_task",
            field=models.CharField(blank=True, max_length=36, null=True),
        ),
        migrations.AlterField(
            model_name="reportdocument",
            name="arguments",
            field=models.JSONField(
                default=dict, encoder=power_query.json.PQJSONEncoder
            ),
        ),
    ]



import time
from typing import Any, Dict, List, Optional

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType

import factory
from factory.django import DjangoModelFactory

from power_query.models import (
    Dataset,
    Formatter,
    Parametrizer,
    Query,
    Report,
    ReportDocument,
)


class UserFactory(DjangoModelFactory):
    class Meta:
        model = get_user_model()
        django_get_or_create = ("username",)

    first_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    email = factory.LazyAttribute(
        lambda o: f"{o.first_name.lower()}.{o.last_name.lower()}_{time.time_ns()}@unicef.com"
    )
    username = factory.LazyAttribute(
        lambda o: f"{o.first_name}{o.last_name}_{time.time_ns()}"
    )

    @classmethod
    def _create(cls, model_class: Any, *args: Any, **kwargs: Any) -> get_user_model():
        manager = cls._get_manager(model_class)
        keyword_arguments = kwargs.copy()
        if "password" not in keyword_arguments:
            keyword_arguments["password"] = "password"
        return manager.create_user(*args, **keyword_arguments)


class GroupFactory(DjangoModelFactory):
    name = factory.Sequence(lambda x: "Group%s" % x)

    class Meta:
        model = Group
        django_get_or_create = ("name",)


class ContentTypeFactory(DjangoModelFactory):
    class Meta:
        model = ContentType
        django_get_or_create = ("app_label", "model")


class QueryFactory(DjangoModelFactory):
    name = factory.Sequence(lambda x: "Query%s" % x)
    owner = factory.SubFactory(UserFactory, is_superuser=True, is_staff=True)
    target = factory.Iterator(
        ContentType.objects.filter(app_label="auth", model="permission")
    )
    code = "result=conn.all()"

    class Meta:
        model = Query
        # django_get_or_create = ("name",)


class FormatterFactory(DjangoModelFactory):
    name = factory.Sequence(lambda x: "Formatter%s" % x)
    content_type = "html"

    class Meta:
        model = Formatter
        django_get_or_create = ("name",)


class ReportFactory(DjangoModelFactory):
    name = factory.Sequence(lambda x: "Report%s" % x)
    query = factory.Iterator(Query.objects.all())
    formatter = factory.Iterator(Formatter.objects.all())
    owner = factory.SubFactory(
        UserFactory, is_superuser=True, is_staff=True, password="123"
    )
    frequence = "mon,tue,wed,thu,fri,sat,sun"

    class Meta:
        model = Report
        django_get_or_create = ("name",)


class ParametrizerFactory(DjangoModelFactory):
    code = "active-business-areas"
    source = None

    class Meta:
        model = Parametrizer
        django_get_or_create = ("code",)


def get_group(
    name: str = "Group1", permissions: Optional[List[Permission]] = None
) -> Group:
    group = GroupFactory(name)
    permission_names = permissions or []
    for permission_name in permission_names:
        try:
            app_label, codename = permission_name.split(".")
        except ValueError:
            raise ValueError("Invalid permission name `{0}`".format(permission_name))
    try:
        permission = Permission.objects.get(
            content_type__app_label=app_label, codename=codename
        )
    except Permission.DoesNotExist:
        raise Permission.DoesNotExist(
            "Permission `{0}` does not exists", permission_name
        )

    group.permissions.add(permission)
    return group


class user_grant_permission:
    def __init__(
        self, user: get_user_model(), permissions: Optional[List[Permission]] = None
    ) -> None:
        self.user = user
        self.permissions = permissions
        self.group: Optional[Group] = None

    def __enter__(self) -> None:
        if hasattr(self.user, "_group_perm_cache"):
            del self.user._group_perm_cache
        if hasattr(self.user, "_officepermissionchecker"):
            del self.user._officepermissionchecker
        if hasattr(self.user, "_perm_cache"):
            del self.user._perm_cache
        self.group = get_group(permissions=self.permissions or [])
        self.user.groups.add(self.group)

    def __exit__(self, e_typ: Any, e_val: Any, trcbak: Any) -> None:
        if all((e_typ, e_val, trcbak)):
            raise e_typ(e_val)
        if self.group:
            self.user.groups.remove(self.group)
            self.group.delete()

    def start(self) -> None:
        """Activate a patch, returning any created mock."""
        self.__enter__()

    def stop(self) -> None:
        """Stop an active patch."""
        return self.__exit__(None, None, None)


class DatasetFactory(DjangoModelFactory):
    query = factory.SubFactory(QueryFactory)

    class Meta:
        model = Dataset
        # django_get_or_create = ("name",)

    @classmethod
    def create(cls, **kwargs: Dict) -> Dataset:
        # ret = super().create(**kwargs)
        q: Query = cls.query.get_factory().create()
        q.run(persist=True)
        return q.datasets.first()


class ReportDocumentFactory(DjangoModelFactory):
    report = factory.SubFactory(ReportFactory)

    class Meta:
        model = ReportDocument

    @classmethod
    def create(cls, **kwargs: Dict) -> ReportDocument:
        fmt = Formatter.objects.get(name="Queryset To HTML")
        r: Report = ReportFactory(query=QueryFactory(), formatter=fmt)
        r.execute(run_query=True)
        return r.documents.first()



# Generated by Django 3.2.15 on 2022-12-06 16:28

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("power_query", "0005_migration"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dataset",
            name="hash",
            field=models.CharField(editable=False, max_length=200, unique=True),
        ),
    ]



from django.test import override_settings, TestCase

from power_query.defaults import create_defaults
from power_query.models import Formatter, Query, Report

from .fixtures import (
    FormatterFactory,
    QueryFactory,
    ReportFactory,
    UserFactory,
)


@override_settings(POWER_QUERY_DB_ALIAS="default")
class TestPowerQuery(TestCase):
    databases = {"default"}

    @classmethod
    def setUpTestData(cls) -> None:
        cls.superuser = UserFactory(is_superuser=True, is_staff=True, is_active=True)
        cls.user1 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        cls.user2 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        create_defaults()
        cls.query1: Query = QueryFactory(name="Query1", code="result=conn.all()")
        cls.query2: Query = QueryFactory(
            name="Query2", code=f"result=invoke({cls.query1.pk}, arguments)"
        )
        cls.formatter: Formatter = FormatterFactory(name="Queryset To HTML")
        cls.report: Report = ReportFactory(formatter=cls.formatter, query=cls.query1)

    def test_query_execution(self) -> None:
        result = self.query1.execute_matrix()
        self.assertTrue(self.query1.datasets.exists())
        self.assertEqual(result["{}"], self.query1.datasets.first().pk)

    def test_query_lazy_execution(self) -> None:
        self.query1.execute_matrix()
        ds1 = self.query1.datasets.first()
        ds2, __ = self.query1.run(use_existing=True)
        self.assertEqual(ds1.pk, ds2.pk)

    def test_report_execution(self) -> None:
        self.query1.execute_matrix()
        dataset = self.query1.datasets.first()
        self.report.execute()
        self.assertTrue(self.report.documents.filter(dataset=dataset).exists())

    def test_nested_query(self) -> None:
        result = self.query2.execute_matrix()
        self.assertTrue(self.query2.datasets.exists())
        self.assertEqual(result["{}"], self.query2.datasets.first().pk)



from django.test import override_settings, TestCase

from power_query.celery_tasks import (
    refresh_report,
    refresh_reports,
    run_background_query,
)
from power_query.defaults import create_defaults
from power_query.models import Formatter, Parametrizer, Query, Report

from .fixtures import (
    FormatterFactory,
    ParametrizerFactory,
    QueryFactory,
    ReportFactory,
    UserFactory,
)


@override_settings(POWER_QUERY_DB_ALIAS="default")
class TestPowerQueryCelery(TestCase):
    databases = {"default"}

    @classmethod
    def setUpTestData(cls) -> None:
        cls.superuser = UserFactory(is_superuser=True, is_staff=True, is_active=True)
        cls.user1 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        cls.user2 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        create_defaults()

        cls.params: Parametrizer = ParametrizerFactory()
        cls.query1: Query = QueryFactory(code="result=conn.all()")
        cls.query2: Query = QueryFactory(
            code=f"result=invoke({cls.query1.pk}, arguments)"
        )
        cls.formatter: Formatter = FormatterFactory(name="Queryset To HTML")
        cls.report: Report = ReportFactory(
            formatter=cls.formatter, query=cls.query1, last_run=None
        )

    def test_query_direct_task(self) -> None:
        target = self.query1.pk
        run_background_query(target)
        assert self.query1.datasets.exists()

    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    def test_query_queue(self) -> None:
        target = self.query1.pk
        res = run_background_query.delay(target)
        assert res
        assert self.query1.datasets.exists()

    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    def test_reports_queue(self) -> None:
        refresh_reports.delay()
        assert self.report.documents.exists()

    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    def test_report_queue(self) -> None:
        target = self.report.pk
        refresh_report.delay(target)
        assert self.report.documents.exists()



# Generated by Django 3.2.18 on 2023-05-31 14:32

from django.db import migrations, models

import power_query.json


class Migration(migrations.Migration):
    dependencies = [
        ("power_query", "0008_migration"),
    ]

    operations = [
        migrations.AlterField(
            model_name="query",
            name="info",
            field=models.JSONField(
                blank=True, default=dict, encoder=power_query.json.PQJSONEncoder
            ),
        ),
    ]



from django.test import TestCase

from freezegun import freeze_time
from parameterized import parameterized

from power_query.utils import should_run


class TestRunToday(TestCase):
    databases = {"default"}

    @parameterized.expand(
        [
            ("Sat", True),
            ("sat", True),
            ("Mon", False),
            ("3", True),
            ("31", False),
            ("3/1", True),
            ("03/01", True),
            ("1/3", False),
            ("1/3,sat", True),
            ("1/3,Mon", False),
            ("1/3,Mon,3/1", True),
            ("mon,tue", False),
            ("mon,sat", True),
            ("3/1,4/1", True),
        ]
    )
    @freeze_time("1987-01-03")
    def test_query_execution(self, pattern: str, value: bool) -> None:
        self.assertEqual(should_run(pattern), value)



# Generated by Django 2.2.16 on 2022-01-18 06:22

import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Formatter",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True, max_length=255, null=True, unique=True
                    ),
                ),
                (
                    "content_type",
                    models.CharField(
                        choices=[
                            ["xls", "application/vnd.ms-excel"],
                            ["txt", "text/plain"],
                            ["csv", "text/csv"],
                            ["html", "text/html"],
                            ["yaml", "text/yaml"],
                            ["json", "application/json"],
                        ],
                        max_length=5,
                    ),
                ),
                ("code", models.TextField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name="Query",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True, max_length=255, null=True, unique=True
                    ),
                ),
                ("description", models.TextField(blank=True, null=True)),
                ("code", models.TextField(blank=True, default="qs=conn.all()")),
                (
                    "info",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True, default=dict
                    ),
                ),
                (
                    "query_args",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True, default=dict
                    ),
                ),
                ("error", models.CharField(blank=True, max_length=400, null=True)),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="power_queries",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "target",
                    models.ForeignKey(
                        default="",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Power Queries",
                "ordering": ("name",),
            },
        ),
        migrations.CreateModel(
            name="Report",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True, max_length=255, null=True, unique=True
                    ),
                ),
                ("refresh", models.BooleanField(default=False)),
                (
                    "query_args",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True, default=dict
                    ),
                ),
                ("last_run", models.DateTimeField(blank=True, null=True)),
                ("result", models.BinaryField(blank=True, null=True)),
                (
                    "available_to",
                    models.ManyToManyField(
                        blank=True,
                        related_name="_report_available_to_+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "formatter",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="power_query.Formatter",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "query",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="power_query.Query",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Dataset",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_run", models.DateTimeField(blank=True, null=True)),
                ("result", models.BinaryField(blank=True, null=True)),
                (
                    "info",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True, default=dict
                    ),
                ),
                (
                    "query",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="power_query.Query",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 2.2.16 on 2022-01-20 09:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("power_query", "0001_migration"),
    ]

    operations = [
        migrations.AlterField(
            model_name="formatter",
            name="content_type",
            field=models.CharField(
                choices=[
                    ["csv", "text/csv"],
                    ["html", "text/html"],
                    ["json", "application/json"],
                    ["txt", "text/plain"],
                    ["xls", "application/vnd.ms-excel"],
                    ["xml", "application/xml"],
                    ["yaml", "text/yaml"],
                ],
                max_length=5,
            ),
        ),
    ]



# Generated by Django 3.2.15 on 2023-01-24 22:10

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("power_query", "0006_migration"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="query",
            name="refresh_daily",
        ),
        migrations.RemoveField(
            model_name="report",
            name="refresh_daily",
        ),
        migrations.AddField(
            model_name="report",
            name="frequence",
            field=models.CharField(
                blank=True,
                help_text="Refresh every (e.g. 3 - 1/3 - mon - 1/3,Mon)",
                max_length=3,
                null=True,
            ),
        ),
    ]



# Generated by Django 3.2.15 on 2022-10-07 08:46

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import power_query.models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("power_query", "0003_migration"),
    ]

    operations = [
        migrations.CreateModel(
            name="Parametrizer",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.SlugField(editable=False, max_length=255, unique=True)),
                ("name", models.CharField(max_length=255, unique=True)),
                (
                    "description",
                    models.TextField(blank=True, max_length=255, null=True),
                ),
                (
                    "value",
                    models.JSONField(
                        default=dict, validators=[power_query.models.validate_queryargs]
                    ),
                ),
                (
                    "system",
                    models.BooleanField(blank=True, default=False, editable=False),
                ),
            ],
            options={
                "verbose_name": "Arguments",
                "verbose_name_plural": "Arguments",
            },
        ),
        migrations.RenameField(
            model_name="dataset",
            old_name="result",
            new_name="value",
        ),
        migrations.RenameField(
            model_name="query",
            old_name="error",
            new_name="sentry_error_id",
        ),
        migrations.RenameField(
            model_name="report",
            old_name="refresh",
            new_name="refresh_daily",
        ),
        migrations.RemoveField(
            model_name="query",
            name="query_args",
        ),
        migrations.RemoveField(
            model_name="report",
            name="available_to",
        ),
        migrations.RemoveField(
            model_name="report",
            name="query_args",
        ),
        migrations.RemoveField(
            model_name="report",
            name="result",
        ),
        migrations.AddField(
            model_name="dataset",
            name="description",
            field=models.CharField(default="", max_length=100),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="dataset",
            name="extra",
            field=models.BinaryField(
                blank=True,
                help_text="Any other attribute to pass to the formatter",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="dataset",
            name="hash",
            field=models.CharField(default="", editable=False, max_length=200),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="query",
            name="active",
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name="query",
            name="error_message",
            field=models.CharField(blank=True, max_length=400, null=True),
        ),
        migrations.AddField(
            model_name="query",
            name="refresh_daily",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="report",
            name="active",
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name="report",
            name="document_title",
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name="report",
            name="limit_access_to",
            field=models.ManyToManyField(
                blank=True,
                related_name="_power_query_report_limit_access_to_+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="dataset",
            name="query",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="datasets",
                to="power_query.query",
            ),
        ),
        migrations.AlterField(
            model_name="query",
            name="name",
            field=models.CharField(max_length=255, unique=True),
        ),
        migrations.AlterField(
            model_name="query",
            name="target",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AddField(
            model_name="query",
            name="parametrizer",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="power_query.parametrizer",
            ),
        ),
        migrations.CreateModel(
            name="ReportDocument",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now=True)),
                ("title", models.CharField(max_length=300)),
                ("output", models.BinaryField(blank=True, null=True)),
                ("arguments", models.JSONField(default=dict)),
                (
                    "content_type",
                    models.CharField(
                        choices=[
                            ["csv", "text/csv"],
                            ["html", "text/html"],
                            ["json", "application/json"],
                            ["txt", "text/plain"],
                            ["xls", "application/vnd.ms-excel"],
                            ["xml", "application/xml"],
                            ["yaml", "text/yaml"],
                        ],
                        max_length=5,
                    ),
                ),
                (
                    "dataset",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="power_query.dataset",
                    ),
                ),
                (
                    "limit_access_to",
                    models.ManyToManyField(
                        blank=True,
                        related_name="_power_query_reportdocument_limit_access_to_+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "report",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="documents",
                        to="power_query.report",
                    ),
                ),
            ],
            options={
                "unique_together": {("report", "dataset")},
            },
        ),
    ]



from django.contrib.auth import get_user_model


def create_defaults() -> None:
    if get_user_model().objects.filter(is_superuser=True).first() is None:
        return
    from power_query.models import Formatter

    Formatter.objects.get_or_create(
        name="Dataset To HTML",
        defaults={
            "code": """
<h1>{{title}}</h1>
<table>
    <tr>{% for fname in dataset.data.headers %}<th>{{ fname }}</th>{% endfor %}</tr>
{% for row in dataset.data %}<tr>{% for col in row %}<td>{{ col }}</td>{% endfor %}</tr>
{% endfor %}
    </table>
"""
        },
    )

    Formatter.objects.get_or_create(
        name="Queryset To HTML",
        defaults={
            "code": """
<h1>{{title}}</h1>
<table>
    <tr><th>id</th><th>str</th></tr>
{% for row in dataset.data %}<tr>
    <td>{{ row.id }}</td>
    <td>{{ row }}</td>
    </tr>
{% endfor %}
    </table>
""",
            "content_type": "html",
        },
    )

    Formatter.objects.get_or_create(
        name="Dataset To XLS", defaults={"code": "", "content_type": "xls"}
    )



# Generated by Django 3.2.18 on 2023-06-01 06:00

from django.db import migrations, models

import power_query.validators


class Migration(migrations.Migration):
    dependencies = [
        ("power_query", "0010_migration"),
    ]

    operations = [
        migrations.AlterField(
            model_name="report",
            name="frequence",
            field=models.CharField(
                blank=True,
                default="mon,tue,wed,thu,fri,sat,sun",
                help_text="Refresh every (e.g. 3 - 1/3 - mon - 1/3,Mon)",
                max_length=30,
                null=True,
                validators=[power_query.validators.FrequencyValidator],
            ),
        ),
    ]



# Generated by Django 3.2.15 on 2022-10-07 08:46

from django.db import migrations

from power_query.utils import dict_hash


def generate_hashes(apps, schema_editor):
    Dataset = apps.get_model("power_query", "Dataset")
    for dataset in Dataset.objects.all():
        arguments = dataset.info.get("arguments", {})
        query_id = dataset.query_id
        dataset.hash = dict_hash({"query": query_id, **arguments})
        dataset.save()


class Migration(migrations.Migration):
    dependencies = [
        ("power_query", "0004_migration"),
    ]

    operations = [
        migrations.RunPython(generate_hashes, migrations.RunPython.noop),
    ]



import logging
from typing import Any, Dict, Tuple, Type, Union

import celery
from billiard.einfo import ExceptionInfo
from celery import Task
from sentry_sdk import capture_exception

from power_query.models import Query, QueryResult, Report, ReportResult
from power_query.utils import sentry_tags, should_run

logger = logging.getLogger(__name__)


class AbstractPowerQueryTask(Task):
    model: Union[Type[Query], Type[Report]]

    def on_success(
        self, retval: Any, task_id: str, args: Tuple[Any], kwargs: Dict[str, Any]
    ) -> None:
        """
        retval (Any): The return value of the task.
        task_id (str): Unique id of the executed task.
        args (Tuple): Original arguments for the executed task.
        kwargs (Dict): Original keyword arguments for the executed task.
        """
        pass

    def on_failure(
        self,
        exc: Exception,
        task_id: str,
        args: Tuple[Any],
        kwargs: Dict[str, Any],
        einfo: ExceptionInfo,
    ) -> None:
        """
        exc (Exception): The exception raised by the task.
        task_id (str): Unique id of the failed task.
        args (Tuple): Original arguments for the task that failed.
        kwargs (Dict): Original keyword arguments for the task that failed.
        einfo (~billiard.einfo.ExceptionInfo): Exception information.
        """
        q = self.model.objects.get(id=args[0])
        sid = capture_exception(exc)
        q.sentry_error_id = sid
        q.error_message = str(exc)
        q.save()


class PowerQueryTask(AbstractPowerQueryTask):
    model = Query


class ReportTask(AbstractPowerQueryTask):
    model = Report


@celery.current_app.task(base=PowerQueryTask)
@sentry_tags
def run_background_query(query_id: int) -> QueryResult:
    try:
        query = Query.objects.get(pk=query_id)
        return query.execute_matrix()
    except BaseException as e:
        logger.exception(e)
        raise


@celery.current_app.task(
    bind=True, default_retry_delay=60, max_retries=3, base=ReportTask
)
@sentry_tags
def refresh_report(self: Any, id: int) -> ReportResult:
    result: ReportResult = []
    try:
        report: Report = Report.objects.get(id=id)
        result = report.execute(run_query=True)
    except Report.DoesNotExist as e:
        logger.exception(e)
    except BaseException as e:
        logger.exception(e)
        raise self.retry(exc=e)
    return result


@celery.current_app.task(
    bind=True, default_retry_delay=60, max_retries=3, base=ReportTask
)
@sentry_tags
def refresh_reports(self: Any) -> Any:
    results: Any = []
    report: Report
    try:
        for report in Report.objects.filter(active=True, frequence__isnull=False):
            if should_run(report.frequence):
                ret = report.queue()
                results.append(ret)
            else:
                results.append([report.pk, "skip"])
    except BaseException as e:
        logger.exception(e)
        raise self.retry(exc=e)
    return results



# Generated by Django 3.2 on 2022-02-11 19:54

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("power_query", "0002_migration"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dataset",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="dataset",
            name="info",
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name="formatter",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="query",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="query",
            name="info",
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name="query",
            name="query_args",
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name="report",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="report",
            name="query_args",
            field=models.JSONField(blank=True, default=dict),
        ),
    ]



# Generated by Django 3.2.18 on 2023-05-19 08:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("power_query", "0012_migration"),
    ]

    operations = [
        migrations.AlterField(
            model_name="formatter",
            name="content_type",
            field=models.CharField(
                choices=[
                    ("csv", "text/csv"),
                    ("html", "text/html"),
                    ("json", "application/json"),
                    ("txt", "text/plain"),
                    ("xls", "application/vnd.ms-excel"),
                    ("xml", "application/xml"),
                    ("yaml", "text/yaml"),
                ],
                max_length=5,
            ),
        ),
        migrations.AlterField(
            model_name="reportdocument",
            name="content_type",
            field=models.CharField(choices=[], max_length=5),
        ),
    ]



from calendar import timegm
from hashlib import md5
from uuid import UUID

from django.contrib.auth.decorators import login_required
from django.core.exceptions import PermissionDenied
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseForbidden,
    JsonResponse,
)
from django.shortcuts import get_object_or_404, render
from django.utils.cache import get_conditional_response
from django.utils.http import http_date

from power_query.models import mimetype_map, Report, ReportDocument
from power_query.utils import basicauth


@login_required()
def report_list(request: HttpRequest) -> HttpResponse:
    return render(request, "power_query/list.html", {"reports": Report.objects.all()})


@login_required()
def report(request: HttpRequest, pk: UUID) -> HttpResponse:
    report: Report = get_object_or_404(Report, pk=pk)
    if request.user.has_perm("power_query.view_report", report):
        if not report.documents.exists():
            return HttpResponse("This report is not currently available", status=400)
        return render(request, "power_query/detail.html", {"report": report})
    else:
        raise PermissionDenied()


@login_required()
def document(request: HttpRequest, report: ReportDocument, pk: UUID) -> HttpResponse:
    doc: ReportDocument = get_object_or_404(ReportDocument, pk=pk, report_id=report)
    res_etag = md5(doc.data.encode()).hexdigest()
    res_last_modified = timegm(doc.timestamp.utctimetuple())

    if not request.user.has_perm("power_query.view_reportdocument", doc):
        return HttpResponseForbidden()

    response = get_conditional_response(
        request,
        etag=res_etag,
        last_modified=res_last_modified,
    )
    if response is None:
        response = HttpResponse(doc.data, content_type=mimetype_map[doc.content_type])
        response.headers["X-PQ-Report"] = doc.report.pk
        response.headers["X-PQ-Document"] = doc.pk
        if doc.content_type == "xls":
            response["Content-Disposition"] = f"attachment; filename={doc.title}.xls"
        else:
            response.headers["Cache-Control"] = "private"
            response.headers["Last-Modified"] = http_date(res_last_modified)
            response.headers["ETag"] = res_etag
    return response


@basicauth
def data(request: HttpRequest, pk: UUID) -> HttpResponse:
    doc: ReportDocument = get_object_or_404(ReportDocument, pk=pk)
    report: Report = doc.report
    if request.user.has_perm("power_query.view_reportdocument", doc):
        if doc.data is None:
            content_types = request.headers.get("Accept", "*/*").split(",")
            if "text/html" in content_types:
                return HttpResponse(
                    "This report is not currently available", status=400
                )
            elif "application/json" in content_types:
                return JsonResponse(
                    {"error": "This report is not currently available"}, status=400
                )
            else:
                return HttpResponse(
                    "This report is not currently available",
                    content_type="text/plain",
                    status=400,
                )
        return HttpResponse(
            doc.data, content_type=report.formatter.get_content_type_display()
        )
    else:
        return HttpResponseForbidden()



from pathlib import Path
from typing import Any, Dict

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.management import BaseCommand, CommandParser


class Command(BaseCommand):
    requires_migrations_checks = False

    def add_arguments(self, parser: CommandParser) -> None:
        subparsers = parser.add_subparsers(
            title="command", dest="command", required=True
        )
        CommandParser(add_help=False)
        test = subparsers.add_parser("test")
        test.add_argument("filename")
        test.add_argument("--target")

        execute = subparsers.add_parser("execute")
        execute.add_argument("id")
        execute.add_argument(
            "--persist",
            action="store_true",
            default=False,
        )

        run = subparsers.add_parser("run")
        run.add_argument("id")
        run.add_argument(
            "--persist",
            action="store_true",
            default=False,
        )
        run.add_argument(
            "--arguments",
            "-a",
            action="store",
            nargs="+",
            default=[],
        )

        subparsers.add_parser("list")

    def _list(self, *args: Any, **options: Any) -> None:
        from power_query.models import Query as PowerQuery

        for q in PowerQuery.objects.all():
            self.stdout.write(f"#{q.id:>5}   {q.name[:30]:<32} {q.last_run}")

    def _test(self, *args: Any, **options: Any) -> None:
        from power_query.models import Query as PowerQuery

        code = Path(options["filename"]).read_text()
        target = options["target"]
        model = ContentType.objects.get_for_model(apps.get_model(target))
        pq = PowerQuery(name="Test", target=model, code=code)
        arguments = {}
        result, info = pq.run(persist=False, arguments=arguments)
        for entry in result:
            print(type(entry), entry)

    def _run(self, *args: Any, **options: Any) -> None:
        from power_query.models import Query as PowerQuery

        query_args: Dict[str, str] = {}
        try:
            for a in options["arguments"]:
                k, v = a.split("=")
                query_args[k] = v
            pq = PowerQuery.objects.get(pk=options["id"])
            result, info = pq.run(persist=options["persist"], arguments=query_args)
            for k, v in info.items():
                self.stdout.write(f"{k}: {v}")
            self.stdout.write("=" * 80)
            for entry in result:
                self.stdout.write(str(entry))
        except Exception as e:
            self.stdout.write(f"Error: {e.__class__.__name__}")
            self.stdout.write(str(e))

    def _execute(self, *args: Any, **options: Any) -> None:
        from power_query.models import Query as PowerQuery

        try:
            pq = PowerQuery.objects.get(pk=options["id"])
            result = pq.execute_matrix(persist=options["persist"])
            for entry in result:
                self.stdout.write(str(entry))
        except Exception as e:
            self.stdout.write(f"Error: {e.__class__.__name__}")
            self.stdout.write(str(e))
            raise

    def handle(self, *args: Any, **options: Any) -> None:
        if options["command"] == "list":
            self._list(*args, **options)
        elif options["command"] == "execute":
            self._execute(*args, **options)
        elif options["command"] == "test":
            self._test(*args, **options)
        elif options["command"] == "run":
            self._run(*args, **options)



from django.test import override_settings, TestCase

from .fixtures import ParametrizerFactory


@override_settings(POWER_QUERY_DB_ALIAS="default")
class TestPowerQuery(TestCase):
    databases = {"default"}

    def test_create_defaults(self) -> None:
        from power_query.defaults import create_defaults

        create_defaults()

    def test_parameter(self) -> None:
        p = ParametrizerFactory()
        p.refresh()



import itertools
import logging
import pickle
from typing import Any, Dict, List, Optional, Tuple, Union

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.db.models import JSONField
from django.template import Context, Template
from django.urls import reverse
from django.utils import timezone
from django.utils.datetime_safe import strftime
from django.utils.functional import cached_property
from django.utils.text import slugify

import celery
from celery import states
from celery.result import AsyncResult
from natural_keys import NaturalKeyModel
from sentry_sdk import capture_exception, configure_scope

from power_query.exceptions import QueryRunError
from power_query.json import PQJSONEncoder
from power_query.utils import dict_hash, to_dataset
from power_query.validators import FrequencyValidator

logger = logging.getLogger(__name__)

mimetype_map = {
    "csv": "text/csv",
    "html": "text/html",
    "json": "application/json",
    "txt": "text/plain",
    "xls": "application/vnd.ms-excel",
    "xml": "application/xml",
    "yaml": "text/yaml",
}

MIMETYPES = ((k, v) for k, v in mimetype_map.items())

DocumentResult = Tuple[int, Union[str, int]]
ReportResult = List[Union[DocumentResult, Any, str]]
QueryResult = Tuple[Any, Any]
QueryMatrixResult = Dict[str, Union[int, str]]


def validate_queryargs(value: Any) -> None:
    try:
        if not isinstance(value, dict):
            raise ValidationError("QueryArgs must be a dict")
        product = list(itertools.product(*value.values()))
        [dict(zip(value.keys(), e)) for e in product]
    except ValidationError:
        raise
    except Exception as e:
        raise ValidationError(
            "%(exc)s: " "%(value)s is not a valid QueryArgs",
            params={"value": value, "exc": e},
        )


class CeleryEnabled(models.Model):
    SCHEDULED = frozenset(
        {states.PENDING, states.RECEIVED, states.STARTED, states.RETRY}
    )

    celery_task = models.CharField(max_length=36, blank=True, null=True)

    class Meta:
        abstract = True

    def status(self) -> str:
        if self.celery_task:
            try:
                result = self.async_result.state
            except Exception as e:
                result = str(e)
        else:
            result = "Not scheduled"
        return result

    @property
    def async_result(self) -> Optional[AsyncResult]:
        if self.celery_task and self.status not in self.SCHEDULED:
            return AsyncResult(self.celery_task, app=celery.current_app)
        else:
            return None

    def queue(self) -> Optional[str]:
        if self.status not in self.SCHEDULED:
            task_id = self._queue()
            if not task_id:
                raise NotImplementedError("`_queue` not properly implemented")
            return task_id
        return None

    def _queue(self) -> str:
        return ""


class Parametrizer(NaturalKeyModel, models.Model):
    code = models.SlugField(max_length=255, unique=True, editable=False)
    name = models.CharField(max_length=255, unique=True)
    description = models.TextField(
        max_length=255,
        null=True,
        blank=True,
    )
    value = models.JSONField(default=dict, blank=False, validators=[validate_queryargs])
    system = models.BooleanField(blank=True, default=False, editable=False)
    source = models.ForeignKey(
        "Query", blank=True, null=True, on_delete=models.CASCADE, related_name="+"
    )

    class Meta:
        verbose_name_plural = "Arguments"
        verbose_name = "Arguments"

    def clean(self) -> None:
        validate_queryargs(self.value)

    def get_matrix(self) -> List[Dict]:
        product = list(itertools.product(*self.value.values()))
        return [dict(zip(self.value.keys(), e)) for e in product]

    def save(
        self,
        force_insert: bool = False,
        force_update: bool = False,
        using: Optional[Any] = None,
        update_fields: Optional[Any] = None,
    ) -> None:
        if not self.code:
            self.code = slugify(self.name)
        super().save(force_insert, force_update, using, update_fields)

    def refresh(self) -> None:
        if self.source:
            out, __ = self.source.run(use_existing=True)
            self.value = out
            self.save()

    def __str__(self) -> str:
        return self.name


class Query(NaturalKeyModel, CeleryEnabled, models.Model):
    name = models.CharField(max_length=255, unique=True)
    description = models.TextField(blank=True, null=True)
    owner = models.ForeignKey(
        get_user_model(), on_delete=models.CASCADE, related_name="power_queries"
    )
    target = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    code = models.TextField(default="qs=conn.all()", blank=True)
    info = JSONField(default=dict, blank=True, encoder=PQJSONEncoder)
    parametrizer = models.ForeignKey(
        Parametrizer, on_delete=models.CASCADE, blank=True, null=True
    )
    sentry_error_id = models.CharField(max_length=400, blank=True, null=True)
    error_message = models.CharField(max_length=400, blank=True, null=True)

    last_run = models.DateTimeField(null=True, blank=True)

    active = models.BooleanField(default=True)

    def __str__(self) -> str:
        return self.name or ""

    class Meta:
        verbose_name_plural = "Power Queries"
        ordering = ("name",)

    def save(
        self,
        force_insert: bool = False,
        force_update: bool = False,
        using: Optional[Any] = None,
        update_fields: Optional[Any] = None,
    ) -> None:
        if not self.code:
            self.code = "qs=conn.all().order_by('id')"
        super().save(force_insert, force_update, using, update_fields)

    def _invoke(self, query_id: int, arguments: List) -> Dict:
        query = Query.objects.get(id=query_id)
        result = query.run(persist=False, arguments=arguments, use_existing=True)
        return result

    def update_results(self, results: QueryMatrixResult) -> None:
        self.info["last_run_results"] = results
        self.error_message = results.get("error_message", "")
        self.sentry_error_id = results.get("sentry_error_id", "")
        self.last_run = timezone.now()
        self.save()

    def execute_matrix(self, persist: bool = True, **kwargs: Any) -> QueryMatrixResult:
        if self.parametrizer:
            args = self.parametrizer.get_matrix()
            if not args:
                raise ValueError("No valid arguments provided")
        else:
            args = [{}]
        self.error_message = None
        self.sentry_error_id = None
        self.last_run = None
        self.info = {}

        results: QueryMatrixResult = {
            "timestamp": strftime(timezone.now(), "%Y-%m-%d %H:%M")
        }
        with configure_scope() as scope:
            scope.set_tag("power_query", True)
            scope.set_tag("power_query.name", self.name)
            with transaction.atomic():
                transaction.on_commit(lambda: self.update_results(results))
                for a in args:
                    try:
                        dataset, __ = self.run(persist, a)
                        if isinstance(dataset, Dataset):
                            results[str(a)] = dataset.pk
                        else:
                            results[str(a)] = str(len(dataset))

                    except QueryRunError as e:
                        logger.exception(e)
                        err = capture_exception(e)
                        results["sentry_error_id"] = str(err)
                        results["error_message"] = str(e)
                self.datasets.exclude(
                    pk__in=[dpk for dpk in results.values() if isinstance(dpk, int)]
                ).delete()
        return results

    def run(
        self,
        persist: bool = False,
        arguments: Optional[Dict] = None,
        use_existing: bool = False,
    ) -> QueryResult:
        model = self.target.model_class()
        connections = {
            f"{model._meta.object_name}Manager": model._default_manager.using(
                settings.POWER_QUERY_DB_ALIAS
            )
            for model in [
                get_user_model(),
            ]
        }
        return_value: QueryResult
        if self.owner.is_superuser:
            connections["QueryManager"] = Query.objects.filter()
        else:
            connections["QueryManager"] = Query.objects.filter(owner=self.owner)

        try:
            locals_ = {
                "conn": model._default_manager.using(settings.POWER_QUERY_DB_ALIAS),
                "query": self,
                "args": arguments,
                "arguments": arguments,
                "invoke": self._invoke,
                **connections,
            }
            signature = dict_hash(
                {"query": self.pk, **(arguments if arguments else {})}
            )
            if use_existing and (
                ds := Dataset.objects.filter(query=self, hash=signature).first()
            ):
                return_value = ds, ds.extra
            else:
                exec(self.code, globals(), locals_)
                result = locals_.get("result", None)
                extra = locals_.get("extra", None)

                if persist:
                    info = {
                        "type": type(result).__name__,
                        "arguments": arguments,
                    }
                    dataset, __ = Dataset.objects.update_or_create(
                        query=self,
                        hash=signature,
                        defaults={
                            "info": info,
                            "last_run": timezone.now(),
                            "value": pickle.dumps(result),
                            "extra": pickle.dumps(extra),
                        },
                    )
                    return_value = dataset, extra
                else:
                    return_value = result, extra
        except Exception as e:
            raise QueryRunError(e) from e
        return return_value

    def _queue(self) -> str:
        from power_query.celery_tasks import run_background_query

        res = run_background_query.delay(self.id)
        self.celery_task = res.id
        self.save()
        return res.id


class Dataset(NaturalKeyModel, models.Model):
    hash = models.CharField(unique=True, max_length=200, editable=False)
    last_run = models.DateTimeField(null=True, blank=True)
    description = models.CharField(max_length=100)
    query = models.ForeignKey(Query, on_delete=models.CASCADE, related_name="datasets")
    value = models.BinaryField(null=True, blank=True)
    info = JSONField(default=dict, blank=True)
    extra = models.BinaryField(
        null=True, blank=True, help_text="Any other attribute to pass to the formatter"
    )

    def __str__(self) -> str:
        return f"Result of {self.query.name} {self.arguments}"

    @property
    def data(self) -> Any:
        return pickle.loads(self.value)

    @property
    def size(self) -> int:
        return len(self.value)

    @property
    def arguments(self) -> Dict:
        return self.info.get("arguments", {})


class Formatter(NaturalKeyModel, models.Model):
    name = models.CharField(max_length=255, blank=True, null=True, unique=True)
    content_type = models.CharField(max_length=5, choices=MIMETYPES)  # type: ignore # internal mypy error
    code = models.TextField(blank=True, null=True)

    def __str__(self) -> str:  # TODO: name is a nullable charfield?
        return self.name or ""

    def render(self, context: Dict) -> str:
        if self.content_type == "xls":
            dt = to_dataset(context["dataset"].data)
            return dt.export("xls")

        if self.code:
            tpl = Template(self.code)
        elif self.content_type == "json":
            dt = to_dataset(context["dataset"].data)
            return dt.export("json")
        elif self.content_type == "yaml":
            dt = to_dataset(context["dataset"].data)
            return dt.export("yaml")
        else:
            raise ValueError("Unable to render")

        return tpl.render(Context(context))


class Report(NaturalKeyModel, CeleryEnabled, models.Model):
    name = models.CharField(max_length=255, blank=True, null=True, unique=True)
    document_title = models.CharField(max_length=255, blank=True, null=True)
    query = models.ForeignKey(Query, on_delete=models.CASCADE)
    formatter = models.ForeignKey(Formatter, on_delete=models.CASCADE)
    active = models.BooleanField(default=True)
    owner = models.ForeignKey(
        get_user_model(),
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="+",
    )
    limit_access_to = models.ManyToManyField(
        get_user_model(), blank=True, related_name="+"
    )
    frequence = models.CharField(
        max_length=30,
        null=True,
        blank=True,
        help_text="Refresh every (e.g. 3 - 1/3 - mon - 1/3,Mon)",
        default="mon,tue,wed,thu,fri,sat,sun",
        validators=[FrequencyValidator()],
    )
    last_run = models.DateTimeField(null=True, blank=True)
    validity_days = models.IntegerField(default=365)

    def save(
        self,
        force_insert: bool = False,
        force_update: bool = False,
        using: Optional[Any] = None,
        update_fields: Optional[Any] = None,
    ) -> None:
        if not self.document_title:
            self.document_title = self.name
        super().save(force_insert, force_update, using, update_fields)

    def execute(self, run_query: bool = False) -> ReportResult:
        # TODO: refactor that
        query: Query = self.query
        result: ReportResult = []
        if run_query:
            query.execute_matrix()
        for dataset in query.datasets.all():
            if not dataset.size:
                continue
            try:
                context = dataset.arguments
                if dataset.extra:
                    context.update(pickle.loads(dataset.extra) or {})

                title = (
                    (self.document_title % context)
                    if self.document_title
                    else self.document_title
                )
                output = self.formatter.render(
                    {
                        "dataset": dataset,
                        "report": self,
                        "title": title,
                        "context": context,
                    }
                )
                res, __ = ReportDocument.objects.update_or_create(
                    report=self,
                    dataset=dataset,
                    defaults={
                        "title": title,
                        "content_type": self.formatter.content_type,
                        "output": pickle.dumps(output),
                        "arguments": dataset.arguments,
                    },
                )
                result.append((res.pk, len(res.output)))
            except Exception as e:
                logger.exception(e)
                result.append((dataset.pk, str(e)))
        self.last_run = timezone.now()
        if not result:
            result = ["No Dataset available"]
        return result

    def __str__(self) -> str:
        return self.name or ""

    def get_absolute_url(self) -> str:
        return reverse("power_query:report", args=[self.pk])

    def _queue(self) -> str:
        from power_query.celery_tasks import refresh_report

        res = refresh_report.delay(self.id)
        self.celery_task = res.id
        self.save()
        return res.id


class ReportDocumentManager(models.Manager):
    def get_queryset(self) -> models.QuerySet:
        return super().get_queryset().select_related("report")


class ReportDocument(models.Model):
    timestamp = models.DateTimeField(auto_now=True)
    title = models.CharField(max_length=300)
    report = models.ForeignKey(
        Report, on_delete=models.CASCADE, related_name="documents"
    )
    dataset = models.ForeignKey(Dataset, on_delete=models.CASCADE)
    output = models.BinaryField(null=True, blank=True)
    arguments = models.JSONField(default=dict, encoder=PQJSONEncoder)
    limit_access_to = models.ManyToManyField(
        get_user_model(), blank=True, related_name="+"
    )
    content_type = models.CharField(max_length=5, choices=MIMETYPES)  # type: ignore # internal mypy error

    objects = ReportDocumentManager()

    class Meta:
        unique_together = ("report", "dataset")

    def __str__(self) -> str:
        return self.title

    @cached_property
    def data(self) -> Any:
        return pickle.loads(self.output)

    @cached_property
    def size(self) -> int:
        return len(self.output)

    def get_absolute_url(self) -> str:
        return reverse("power_query:document", args=[self.report.pk, self.pk])



# Generated by Django 3.2.19 on 2023-06-10 12:29

import django.db.migrations.operations.special
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import power_query.json
import power_query.models
import power_query.validators
from power_query.utils import dict_hash


def generate_hashes(apps, schema_editor):
    Dataset = apps.get_model("power_query", "Dataset")
    for dataset in Dataset.objects.all():
        arguments = dataset.info.get("arguments", {})
        query_id = dataset.query_id
        dataset.hash = dict_hash({"query": query_id, **arguments})
        dataset.save()


class Migration(migrations.Migration):
    replaces = [
        ("power_query", "0001_migration"),
        ("power_query", "0002_migration"),
        ("power_query", "0003_migration"),
        ("power_query", "0004_migration"),
        ("power_query", "0005_migration"),
        ("power_query", "0006_migration"),
        ("power_query", "0007_migration"),
        ("power_query", "0008_migration"),
        ("power_query", "0009_migration"),
        ("power_query", "0010_migration"),
        ("power_query", "0011_migration"),
        ("power_query", "0012_migration"),
        ("power_query", "0013_migration"),
        ("power_query", "0014_migration"),
    ]

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Formatter",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True, max_length=255, null=True, unique=True
                    ),
                ),
                (
                    "content_type",
                    models.CharField(
                        choices=[
                            ["csv", "text/csv"],
                            ["html", "text/html"],
                            ["json", "application/json"],
                            ["txt", "text/plain"],
                            ["xls", "application/vnd.ms-excel"],
                            ["xml", "application/xml"],
                            ["yaml", "text/yaml"],
                        ],
                        max_length=5,
                    ),
                ),
                ("code", models.TextField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name="Query",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, unique=True)),
                ("description", models.TextField(blank=True, null=True)),
                ("code", models.TextField(blank=True, default="qs=conn.all()")),
                ("info", models.JSONField(blank=True, default=dict)),
                (
                    "sentry_error_id",
                    models.CharField(blank=True, max_length=400, null=True),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="power_queries",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "target",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                (
                    "error_message",
                    models.CharField(blank=True, max_length=400, null=True),
                ),
                ("refresh_daily", models.BooleanField(default=False)),
            ],
            options={
                "verbose_name_plural": "Power Queries",
                "ordering": ("name",),
            },
        ),
        migrations.CreateModel(
            name="Parametrizer",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.SlugField(editable=False, max_length=255, unique=True)),
                ("name", models.CharField(max_length=255, unique=True)),
                (
                    "description",
                    models.TextField(blank=True, max_length=255, null=True),
                ),
                (
                    "value",
                    models.JSONField(
                        default=dict, validators=[power_query.models.validate_queryargs]
                    ),
                ),
                (
                    "system",
                    models.BooleanField(blank=True, default=False, editable=False),
                ),
            ],
            options={
                "verbose_name": "Arguments",
                "verbose_name_plural": "Arguments",
            },
        ),
        migrations.CreateModel(
            name="Report",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True, max_length=255, null=True, unique=True
                    ),
                ),
                ("refresh_daily", models.BooleanField(default=False)),
                ("last_run", models.DateTimeField(blank=True, null=True)),
                (
                    "formatter",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="power_query.formatter",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "query",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="power_query.query",
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                (
                    "document_title",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                (
                    "limit_access_to",
                    models.ManyToManyField(
                        blank=True,
                        related_name="_power_query_report_limit_access_to_+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Dataset",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_run", models.DateTimeField(blank=True, null=True)),
                ("value", models.BinaryField(blank=True, null=True)),
                ("info", models.JSONField(blank=True, default=dict)),
                (
                    "query",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="datasets",
                        to="power_query.query",
                    ),
                ),
                ("description", models.CharField(default="", max_length=100)),
                (
                    "extra",
                    models.BinaryField(
                        blank=True,
                        help_text="Any other attribute to pass to the formatter",
                        null=True,
                    ),
                ),
                ("hash", models.CharField(default="", editable=False, max_length=200)),
            ],
        ),
        migrations.AddField(
            model_name="query",
            name="parametrizer",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="power_query.parametrizer",
            ),
        ),
        migrations.CreateModel(
            name="ReportDocument",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now=True)),
                ("title", models.CharField(max_length=300)),
                ("output", models.BinaryField(blank=True, null=True)),
                ("arguments", models.JSONField(default=dict)),
                (
                    "content_type",
                    models.CharField(
                        choices=[
                            ["csv", "text/csv"],
                            ["html", "text/html"],
                            ["json", "application/json"],
                            ["txt", "text/plain"],
                            ["xls", "application/vnd.ms-excel"],
                            ["xml", "application/xml"],
                            ["yaml", "text/yaml"],
                        ],
                        max_length=5,
                    ),
                ),
                (
                    "dataset",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="power_query.dataset",
                    ),
                ),
                (
                    "limit_access_to",
                    models.ManyToManyField(
                        blank=True,
                        related_name="_power_query_reportdocument_limit_access_to_+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "report",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="documents",
                        to="power_query.report",
                    ),
                ),
            ],
            options={
                "unique_together": {("report", "dataset")},
            },
        ),
        migrations.RunPython(generate_hashes, migrations.RunPython.noop),
        migrations.AlterField(
            model_name="dataset",
            name="hash",
            field=models.CharField(editable=False, max_length=200, unique=True),
        ),
        migrations.RemoveField(
            model_name="query",
            name="refresh_daily",
        ),
        migrations.RemoveField(
            model_name="report",
            name="refresh_daily",
        ),
        migrations.AddField(
            model_name="query",
            name="last_run",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="formatter",
            name="content_type",
            field=models.CharField(
                choices=[
                    ("csv", "text/csv"),
                    ("html", "text/html"),
                    ("json", "application/json"),
                    ("txt", "text/plain"),
                    ("xls", "application/vnd.ms-excel"),
                    ("xml", "application/xml"),
                    ("yaml", "text/yaml"),
                ],
                max_length=5,
            ),
        ),
        migrations.AlterField(
            model_name="reportdocument",
            name="content_type",
            field=models.CharField(choices=[], max_length=5),
        ),
        migrations.AlterField(
            model_name="query",
            name="info",
            field=models.JSONField(
                blank=True, default=dict, encoder=power_query.json.PQJSONEncoder
            ),
        ),
        migrations.AddField(
            model_name="report",
            name="validity_days",
            field=models.IntegerField(default=365),
        ),
        migrations.AddField(
            model_name="report",
            name="frequence",
            field=models.CharField(
                blank=True,
                default="mon,tue,wed,thu,fri,sat,sun",
                help_text="Refresh every (e.g. 3 - 1/3 - mon - 1/3,Mon)",
                max_length=30,
                null=True,
                validators=[power_query.validators.FrequencyValidator()],
            ),
        ),
        migrations.AlterField(
            model_name="formatter",
            name="content_type",
            field=models.CharField(
                choices=[
                    ("csv", "text/csv"),
                    ("html", "text/html"),
                    ("json", "application/json"),
                    ("txt", "text/plain"),
                    ("xls", "application/vnd.ms-excel"),
                    ("xml", "application/xml"),
                    ("yaml", "text/yaml"),
                ],
                max_length=5,
            ),
        ),
        migrations.AlterField(
            model_name="reportdocument",
            name="content_type",
            field=models.CharField(choices=[], max_length=5),
        ),
        migrations.AddField(
            model_name="parametrizer",
            name="source",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="power_query.query",
            ),
        ),
        migrations.AddField(
            model_name="query",
            name="celery_task",
            field=models.CharField(blank=True, max_length=36, null=True),
        ),
        migrations.AddField(
            model_name="report",
            name="celery_task",
            field=models.CharField(blank=True, max_length=36, null=True),
        ),
        migrations.AlterField(
            model_name="reportdocument",
            name="arguments",
            field=models.JSONField(
                default=dict, encoder=power_query.json.PQJSONEncoder
            ),
        ),
    ]
